//use super::util::basicrenderable::BasicRenderable;
//use crate::enginesettings::EngineSettings;
use crate::prelude::*;

use anyhow::Result;
use std::marker::PhantomData;
use std::sync::{
    atomic::{AtomicUsize, Ordering::SeqCst},
    Arc, Mutex,
};

pub struct ComputeShaderUnit<'a> {
    pub descriptor_layouts: &'a [&'a dyn VkHandle<VkDescriptorSetLayout>],
    pub push_constants: &'a [VkPushConstantRange],
    pub shader_module: &'a Arc<ShaderModule<shader_type::Compute>>,
}

impl<'a> ComputeShaderUnit<'a> {
    pub fn new(
        descriptor_layouts: &'a [&'a dyn VkHandle<VkDescriptorSetLayout>],
        push_constants: &'a [VkPushConstantRange],
        shader_module: &'a Arc<ShaderModule<shader_type::Compute>>,
    ) -> Self {
        ComputeShaderUnit {
            descriptor_layouts,
            push_constants,
            shader_module,
        }
    }
}

pub struct SafeCommandBuffer<T> {
    phantom: PhantomData<T>,
    pipelines: Vec<Arc<Pipeline>>,
    renderpass: Option<Arc<RenderPass>>,
    sub_pass_index: Option<u32>,
    command_buffers: Arc<ValidCommandBuffers>,

    pipeline_index: AtomicUsize,
}

impl<T: VertexInputDescription + ReprC + Send + Sync + 'static> SafeCommandBuffer<T> {
    #[allow(clippy::too_many_arguments)]
    pub fn new_rasterize<'a>(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        image_count: usize,
        renderpass: &Arc<RenderPass>,
        sub_pass_index: u32,
        pipeline: Arc<Pipeline>,
    ) -> Result<SafeCommandBuffer<T>> {
        Self::_new(
            device,
            queue,
            image_count,
            vec![pipeline],
            Some(sub_pass_index),
            Some(renderpass.clone()),
        )
    }

    pub fn render(
        &self,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        descriptor_sets: &[&Arc<DescriptorSet>],
        vertex_buffer: &Arc<Buffer<T>>,
    ) -> Result<()> {
        buffer_recorder.bind_descriptor_sets_minimal(descriptor_sets);
        buffer_recorder.bind_vertex_buffer(vertex_buffer);
        buffer_recorder.draw_complete_single_instance(vertex_buffer.size() as u32);

        Ok(())
    }
}

impl<T> SafeCommandBuffer<T> {
    pub fn new_compute(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        image_count: usize,
        pipeline_infos: &[ComputeShaderUnit<'_>],
    ) -> Result<SafeCommandBuffer<T>> {
        let mut pipelines = Vec::new();

        for pipeline_info in pipeline_infos {
            let mut pipeline_layout_builder = PipelineLayout::builder();

            for set_layout in pipeline_info.descriptor_layouts.iter() {
                pipeline_layout_builder =
                    pipeline_layout_builder.add_descriptor_set_layout(*set_layout);
            }

            for push_constant in pipeline_info.push_constants.iter() {
                pipeline_layout_builder =
                    pipeline_layout_builder.add_push_constant(push_constant.clone());
            }

            pipelines.push(
                Pipeline::new_compute()
                    .set_shader_module(pipeline_info.shader_module)
                    .build(device, &pipeline_layout_builder.build(device.clone())?)?,
            );
        }

        Self::_new(device, queue, image_count, pipelines, None, None)
    }

    fn _new(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        image_count: usize,
        pipelines: Vec<Arc<Pipeline>>,
        sub_pass_index: Option<u32>,
        render_pass: Option<Arc<RenderPass>>,
    ) -> Result<SafeCommandBuffer<T>> {
        Ok(SafeCommandBuffer {
            phantom: PhantomData,

            pipelines,
            sub_pass_index,
            renderpass: render_pass,

            command_buffers: ValidCommandBuffers::new(device, queue, image_count)?,

            pipeline_index: AtomicUsize::new(0),
        })
    }

    pub fn invalidate(&self) {
        self.command_buffers.reset_all();
    }

    pub fn current(&self, index: usize) -> Option<&Arc<CommandBuffer>> {
        let (command_buffer, valid) = self.command_buffers.current(index);

        if valid.load(SeqCst) {
            Some(command_buffer)
        } else {
            None
        }
    }

    pub fn begin(
        &self,
        framebuffer: Option<&Arc<Framebuffer>>,
        index: usize,
    ) -> Result<CommandBufferRecorder<'_>> {
        let (command_buffer, _) = self.command_buffers.current(index);

        let inheritance_info = CommandBuffer::inheritance_info(
            self.renderpass.as_ref(),
            self.sub_pass_index,
            framebuffer,
            None,
        );

        let mut begin_info = VkCommandBufferBeginInfo::new(
            VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT
                | VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT,
        );

        begin_info.set_inheritance_info(&inheritance_info);

        let mut buffer_recorder = command_buffer.begin(begin_info)?;

        self.pipeline_index.store(0, SeqCst);
        buffer_recorder.bind_pipeline(&self.pipelines[self.pipeline_index.load(SeqCst)])?;

        Ok(buffer_recorder)
    }

    pub fn begin_headless(&self, index: usize) -> Result<CommandBufferRecorder<'_>> {
        let (command_buffer, _) = self.command_buffers.current(index);

        let inheritance_info = CommandBuffer::inheritance_info(None, None, None, None);

        let mut begin_info =
            VkCommandBufferBeginInfo::new(VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT);

        begin_info.set_inheritance_info(&inheritance_info);

        let mut buffer_recorder = command_buffer.begin(begin_info)?;

        self.pipeline_index.store(0, SeqCst);
        buffer_recorder.bind_pipeline(&self.pipelines[self.pipeline_index.load(SeqCst)])?;

        Ok(buffer_recorder)
    }

    pub fn next_pipeline(&self, buffer_recorder: &mut CommandBufferRecorder<'_>) -> Result<()> {
        assert!(
            self.pipeline_index.load(SeqCst) >= self.pipelines.len() - 1,
            "requested pipeline that does not exist"
        );

        self.pipeline_index
            .store(self.pipeline_index.load(SeqCst) + 1, SeqCst);

        buffer_recorder.bind_pipeline(&self.pipelines[self.pipeline_index.load(SeqCst)])?;

        Ok(())
    }

    pub fn current_pipeline(&self) -> &Arc<Pipeline> {
        &self.pipelines[self.pipeline_index.load(SeqCst)]
    }

    pub fn end(
        &self,
        _buffer_recorder: CommandBufferRecorder<'_>,
        index: usize,
    ) -> Result<&Arc<CommandBuffer>> {
        let (command_buffer, valid) = self.command_buffers.current(index);

        valid.store(true, SeqCst);

        // consuming buffer recorder and drop to end recording

        Ok(command_buffer)
    }

    pub fn dispatch<P>(
        &self,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        descriptor_sets: &[&Arc<DescriptorSet>],
        push_data: Option<P>,
        vertex_count: u32,
    ) -> Result<()> {
        if let Some(push_data) = push_data {
            buffer_recorder.push_constants(VK_SHADER_STAGE_COMPUTE_BIT, &push_data);
        }

        buffer_recorder.bind_descriptor_sets_minimal(descriptor_sets);
        buffer_recorder.dispatch(vertex_count, 1, 1);

        Ok(())
    }
}
